/*  Copyright 2004-2012, Unpublished Work of Technologic Systems
 *  All Rights Reserved.
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR
 *  ASSIGNMENTS  AND (II) ENTITIES OTHER THAN TECHNOLOGIC SYSTEMS WHO
 *  HAVE ENTERED INTO  APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS
 *  WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED,
 *  MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED,
 *  COMPILED,LINKED,RECAST, TRANSFORMED, ADAPTED IN ANY FORM OR BY ANY
 *  MEANS,MANUAL, MECHANICAL, CHEMICAL, ELECTRICAL, ELECTRONIC, OPTICAL,
 *  BIOLOGICAL, OR OTHERWISE WITHOUT THE PRIOR WRITTEN PERMISSION AND
 *  CONSENT OF TECHNOLOGIC SYSTEMS . ANY USE OR EXPLOITATION OF THIS WORK
 *  WITHOUT THE PRIOR WRITTEN CONSENT OF TECHNOLOGIC SYSTEMS  COULD
 *  SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 */
#include <assert.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <linux/unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <limits.h>
#include <sched.h>
#include <libgen.h>
#include <time.h>
#include <sys/timex.h>

#define peek32(adr) syscon[(adr)/4]
#define poke32(adr, dat) syscon[(adr)/4] = (dat)

// See Xilinx UG380 v2.3 page 126
// This bitstream is just a command to load the other bitstream from address
// 0x700000 of the flash.
unsigned long bitstream[] = {
   
  0xffff, 0xffff, 0xffff, 0xffff,
  0xffff, 0xffff, 0xffff, 0xffff,
  0xffff, 0xffff, 0xffff, 0xffff,
  
  0xffff, 0xaa99, 0x5566, 0x3261,
  0x0000, 0x3281, 0x6b70, 0x32a1,
  0x0000, 0x32c1, 0x6b70, 0x30a1,
  0x000e, 0x2000, 0x2000, 0x2000,
  0x2000, 0x2000 //, 0x2000, 0x2000
};

#define NWORDS  (sizeof(bitstream) / sizeof(unsigned long))

static void usage(char **argv) {
	fprintf(stderr, "Usage: %s [OPTION] ...\n"
	  "Technologic Systems TS-4300 control utility.\n"
	  "\n"
	  "General options:\n"
	  "  -i, --info              Display board FPGA info\n"
	  "  -B, --baseboard         Display baseboard ID\n"
	  "  -W, --watchdog          Daemonize and set up /dev/watchdog\n"
	  "  -A, --autofeed=SETTING  Daemonize and auto feed watchdog\n"
	  "  -l, --loadfpga          Load FPGA bitstream from flash\n"
	  "  -M, --mico32=FILE       Load FILE into Lattice Mico32 RAM and execute\n"
	  "  -c, --console           Connect to LM32 output FIFO\n"
	  "  -t, --getrtc            Display RTC time/date\n"
	  "  -S, --setrtc            Set RTC time/date from system time\n"
	  "  -F, --diffrtc           Print RTC drift from system time\n"
	  "  -h, --help              This help\n",
	  argv[0]
	);
}

static inline
unsigned char tobcd(unsigned int n) {
	unsigned char ret;
	ret = (n / 10) << 4;
	ret |= n % 10;
	return ret;
}

static inline unsigned int frombcd(unsigned char x) {
	unsigned int ret;
	ret = (x >> 4) * 10;
	ret += x & 0xf;
	return ret;
}

int main(int argc, char **argv) {
	int devmem, i, c, wdog = 0;
	unsigned long hdr[16];
	unsigned long bridge[16];
	unsigned int baseboard = 0;
	int opt_getrtc = 0, opt_setrtc = 0, opt_info = 0, opt_loadfpga = 0;
	int opt_watchdog = 0, opt_autofeed = 0, opt_baseboard = 0;
	int opt_diffrtc = 0, opt_console = 0;
	char *opt_lm32 = NULL;
	static volatile char *lmfifo;
	volatile unsigned int *syscon;
	static volatile unsigned long *cns3xxx_pcie0_host_base;
	static volatile unsigned long *cns3xxx_pcie0_cfg0_base;
	static volatile unsigned long *misc_control_register_base;
	static volatile unsigned long *cns3xxx_gpioa_base;
	static volatile unsigned long *cns3xxx_gpiob_base;
	FILE *ifile = NULL;

	static struct option long_options[] = {
	  { "help", 0, 0, 'h' },
	  { "getrtc", 0, 0, 't' },
	  { "setrtc", 0, 0, 'S' },
	  { "diffrtc", 0, 0, 'F' },
	  { "info", 0, 0, 'i' },
	  { "watchdog", 0, 0, 'W' },
	  { "autofeed", 1, 0, 'A' },
	  { "baseboard", 0, 0, 'B' },
	  { "loadfpga", 0, 0, 'l' },
	  { "mico32", 1, 0, 'M' },
	  { "console", 0, 0, 'c' },
	  { 0, 0, 0, 0 }
	};

	while((c = getopt_long(argc, argv,
	  "htSiWA:BlFM:c", long_options, NULL)) != -1) {
		switch (c) {
		case 'l':
			opt_loadfpga = 1;
			break;
		case 'M':
			opt_lm32 = strdup(optarg);
			if (strcmp("-", optarg) == 0) ifile = stdin;
			else ifile = fopen(optarg, "r");
			if (!ifile) {
				perror(optarg);
				return 3;
			}
			break;
		case 'c':
			opt_console = 1;
			break;
		case 'W':
			opt_watchdog = 1;
			break;
		case 'A':
			opt_watchdog = 1;
			opt_autofeed = 1;
			wdog = strtoul(optarg, NULL, 0);
			if (wdog > 3) {
				fprintf(stderr, 
				  "autofeed must be between 0 and 3\n");
				exit(1);
			}
			break;
		case 'i':
			opt_info = 1;
			break;
		case 'B':
			opt_baseboard = 1;
			break;
		case 'S':
			opt_setrtc = 1;
			break;
		case 't':
			opt_getrtc = 1;
			break;
		case 'F':
			opt_diffrtc = 1;
			break;
		case 'h':
		default:
			usage(argv);
			return(1);
		}
	} 
	
	devmem = open("/dev/mem", O_RDWR|O_SYNC);
	assert(devmem != -1);
	syscon = (unsigned int *) mmap(0, 4096,
	  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0xa0000000);
	lmfifo = (char *)syscon + 0x800;

	cns3xxx_gpiob_base = mmap(0, getpagesize(), PROT_READ |
	  PROT_WRITE, MAP_SHARED, devmem, 0x74800000);
	if (cns3xxx_gpiob_base == MAP_FAILED) {
		perror("mmap:");
		return 0;
	}

#define MISC_CONTROL_REGISTER_BASE 0x76000000
	misc_control_register_base = mmap(0, getpagesize(), PROT_READ |
	  PROT_WRITE, MAP_SHARED, devmem, MISC_CONTROL_REGISTER_BASE);
	if (misc_control_register_base == MAP_FAILED) {
		perror("mmap:");
		return 0;
	}

#define CNS3XXX_PCIE0_HOST_BASE 0xAB000000 /* PCIe Port 0 RC Base      */
#define CNS3XXX_PCIE0_CFG0_BASE 0xAD000000 /* PCIe Port 0 CFG Type 0   */
	if (opt_loadfpga) {

		cns3xxx_pcie0_host_base =  mmap(0, getpagesize(), PROT_READ |
		  PROT_WRITE, MAP_SHARED, devmem, CNS3XXX_PCIE0_HOST_BASE);
		if (cns3xxx_pcie0_host_base == MAP_FAILED) {
			perror("mmap:");
			return 0;
		}

		cns3xxx_pcie0_cfg0_base = mmap(0, getpagesize(), PROT_READ |
		  PROT_WRITE, MAP_SHARED, devmem, CNS3XXX_PCIE0_CFG0_BASE);
		if (cns3xxx_pcie0_cfg0_base == MAP_FAILED) {
			perror("mmap:");
			return 0;
		}

		cns3xxx_gpioa_base = mmap(0, getpagesize(), PROT_READ |
		  PROT_WRITE, MAP_SHARED, devmem, 0x74000000);
		if (cns3xxx_gpioa_base == MAP_FAILED) {
			perror("mmap:");
			return 0;
		}

		// DONE pin to GPIO functionality:
		misc_control_register_base[0x14 / 4] &= ~(1 << 29);

		/* Save bridge config space */
		for (i = 0; i < 16; i++) {
			bridge[i] = cns3xxx_pcie0_host_base[i];
		}

		/* Save CFG0 space */
		for (i = 0; i < 16; i++) {
			hdr[i] = cns3xxx_pcie0_cfg0_base[i];
		}

		/* Set up the FPGA reprogram */
		syscon[8] |= 0x20000;
		syscon[8] = 0x30000 | bitstream[0];
 
		for (i = 1; i < (NWORDS-1); i++) {
			syscon[8] = bitstream[i] | 0x10000;
			syscon[8] = bitstream[i] | 0x00000;
			syscon[8] = bitstream[i] | 0x10000;
		}

		syscon[8] = bitstream[i-1] | 0x30000;

		usleep(50000); // plenty of time for programming to start
		// monitor DONE pin:
		while (!(cns3xxx_gpioa_base[1] & (1 << 29))) usleep(1);

		//printf("PCIe Link is %s\n", 
		  //(misc_control_register_base[0x960 / 4] & 0x1)?"up":"down");
		assert((misc_control_register_base[0x960 / 4] & 0x1));

		/* Perform bridge reset */
		cns3xxx_pcie0_host_base[0x3E / 4] |= 0x40;
		usleep(20000);
		cns3xxx_pcie0_host_base[0x3E / 4] &= ~0x40;
		usleep(20000);
		cns3xxx_pcie0_host_base[0x3E / 4] &= ~0x40;
		usleep(20000);

		/* Restore CFG0 space */
		for (i = 0; i < 16; i++) {                
			if (i == 1) continue;
			cns3xxx_pcie0_cfg0_base[i] = hdr[i];
		}
		cns3xxx_pcie0_cfg0_base[1] = hdr[1];
		
		/* Restore bridge space */
		for (i = 0; i < 16; i++) {
			if (i == 1) continue;
			cns3xxx_pcie0_host_base[i] = bridge[i];
		}
		cns3xxx_pcie0_host_base[1] = bridge[1];

		munmap((void*)cns3xxx_pcie0_cfg0_base, getpagesize());
		munmap((void*)cns3xxx_pcie0_host_base, getpagesize());
	}

	if (opt_lm32) {
		unsigned int adr;
		volatile unsigned int *shared;
		volatile char *blockram;
		char *buf, *bufpg;
		unsigned int t1, t2;
		int ret;

		shared = (unsigned int *) mmap(0, 4096,
		  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0xa0012000);

		t1 = syscon[5];
		shared[0] = t1;

		buf = malloc(0x2000);
		bufpg = (char *)(((unsigned int)buf + 0xfff) & ~0xfff);

		syscon[1] |= 0x800000; // put LM32 in reset
		adr = 0xa0040000;
		while (!feof(ifile)) {
			blockram = mmap(0, 4096,
			  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, adr);
			ret = fread(bufpg, 1, 0x1000, ifile);
			memcpy((char *)blockram, bufpg, ret);
			munmap((void *)blockram, 0x1000);
			adr += 0x1000;
		}
		t2 = syscon[5];
		syscon[1] &= ~0x800000; // take LM32 out of reset
		printf("lm32_load_usec=%d\n", t2 - t1);
		while(shared[0] == t1) {}; // wait for LM32 to overwrite
		printf("lm32_init_usec=%d\n", shared[0] - t2);
		munmap((void *)shared, 0x1000);
	}

	if (opt_console) while (1) {
		i = lmfifo[1];
		if (i & 0x80) usleep(1);
		else {
			if (i & 0x60) i = 31;
			else i = i & 0x1f;
			while(i--) putchar(lmfifo[0]);
			fflush(stdout);
		}
	}

	if (opt_info) {
		unsigned int x;
		x = peek32(0);
		printf("model=0x%x\n", x & 0xffff);
		printf("submodel=0x%x\n", (x >> 24) & 0xf);
		printf("revision=0x%x\n", (x >> 16) & 0xff);
		printf("bootdev=0x%x\n", (x >> 28) & 0x1);
		//printf("latched_dio9=0x%x\n", (x >> 29) & 0x1); // XXX
		printf("resetsw_en=0x%x\n", (peek32(4) >> 30) & 0x1);
		opt_baseboard = 1;
	}

	if (opt_baseboard) {
		unsigned int prev1, prev2, prev3, prev4, prev5;
		unsigned int i, x;

		// red LED, green LED, and DIO 93 are the outputs
		// DIO 5 is the input
		prev1 = peek32(0x10); // MUXBUS config
		prev2 = peek32(0x48); // DIO DATA
		prev3 = peek32(0x58); // DIO DDR
		prev4 = peek32(0x50); // DIO DDR
		prev5 = peek32(0x04); // control register
		poke32(0x10, 0);
		poke32(0x50, prev4 & ~0x20);
		poke32(0x58, prev3 | 0x20000000);
		for(i=0; i<8; i++) {
			x = prev2 & ~0x20000000;
			if ((i & 4)) x |= 0x20000000;
			poke32(0x48, x);
			x = prev5 & ~0x3;
			if (!(i & 1)) x |= 0x2;
			if (!(i & 2)) x |= 0x1;
			poke32(0x4, x);
			usleep(1);
			baseboard = (baseboard >> 1);
			if (peek32(0x60) & 0x20) baseboard |= 0x80;
		}
		poke32(0x10, prev1);
		poke32(0x48, prev2);
		poke32(0x58, prev3);
		poke32(0x50, prev4);
		poke32(0x04, prev5);
		printf("baseboard_model=0x%x\n", baseboard);
		switch (baseboard) {
			case 2:
				printf("baseboard=8390\n");
				break;
			case 4:
				printf("baseboard=8500\n");
				break;
			case 6:
				printf("baseboard=8160\n");
				break;
			case 7:
				printf("baseboard=8100\n");
				break;
			case 8:
				printf("baseboard=8820\n");
				break;
			case 10:
				printf("baseboard=8900\n");
				break;
			case 11:
				printf("baseboard=8290\n");
				break;
			case 14:
				printf("baseboard=8280\n");
				break;
			case 255:
				printf("baseboard=8200\n");
				break;
		}

	}


	if (opt_setrtc || opt_getrtc || opt_diffrtc) {
		misc_control_register_base[0x18 / 4] &= ~0x00003000; /* pins to GPIO */
                cns3xxx_gpiob_base[0x8 / 4] &= ~0x00003000; /* pins high-Z */
                cns3xxx_gpiob_base[0x14 / 4] = 0x00003000; /* 0 for outputs */
        }

/* I2C bit-banging macros:
 * Output registers will be initialized to zero, and both lines
 * will be toggled only by changing the direction register.
 * This definition of I2C_pause conveniently results in the bus running
 * right around 400kHz where it should.  With no pause, it runs at about
 * 800kHz and it still works, but since the RTC datasheet specifies 400kHz
 * max, I am using the minimal pause.  If we want to be really paranoid we
 * can change the 1 to a 2, but this will reduce the accuracy of the RTC
 * drift measurements.
 */
#define I2C_pause for(k=0; k < 1; k++) dummy = peek32(0)
#define scl_z I2C_pause; cns3xxx_gpiob_base[0x8/4] &= ~(1 << 12)
#define scl_0 I2C_pause; cns3xxx_gpiob_base[0x8/4] |= (1 << 12)
#define sda_z I2C_pause; cns3xxx_gpiob_base[0x8/4] &= ~(1 << 13)
#define sda_0 I2C_pause; cns3xxx_gpiob_base[0x8/4] |= (1 << 13)
#define sda_in (cns3xxx_gpiob_base[0x4/4] & (1 << 13))

	if (opt_setrtc) {
		unsigned int j, k, d, ack;
		unsigned char rtcdat[10];
		time_t now = time(NULL);
		struct tm *tm;
		//int ppb;
                volatile unsigned int dummy;

	//	read_tagmem(tagmem);
	//	ppb = (signed char)(~tagmem[2] & 0xff) * 1000;
	//	if (ppb > (4068 * 31)) ppb = (4068 * 31);
	//	if (ppb < (-2034 * 31)) ppb = (-2034 * 31);
	//	if (ppb >= 0) {
	//		ppb = ppb / 4068;
	//		rtcdat[8] = (1<<5)|(ppb & 0x1f);
	//	} else {
	//		ppb = ppb / -2034;
	//		rtcdat[8] = (ppb & 0x1f);
	//	}
		now += 6; /* 6 seconds in future to accomodate sleep() below */
		fprintf(stderr, "rtc_calibration_reg=0x%x\n", rtcdat[8]);
		fprintf(stderr, "rtc_calibration_ppm=%f\n", rtcdat[8]&(1<<5) ?
		  ((rtcdat[8] & 0x1f) * 4.068) : ((rtcdat[8] & 0x1f) * -2.034));
		tm = gmtime(&now);
		rtcdat[0] = 0;
		rtcdat[1] = tobcd(tm->tm_sec) | 0x80;
		rtcdat[2] = tobcd(tm->tm_min) | 0x80;
		rtcdat[3] = tobcd(tm->tm_hour);
		rtcdat[4] = tm->tm_wday + 1; 
		rtcdat[5] = tobcd(tm->tm_mday);
		rtcdat[6] = tobcd(tm->tm_mon + 1);
		assert(tm->tm_year >= 100);
		rtcdat[7] = tobcd(tm->tm_year % 100);

resettime:
		scl_z; /* tristate, scl/sda pulled up */
                sda_z;
		sda_0; /* i2c, start (sda low) */
		for (d = 0xd0, i = 0; i < 8; i++, d <<= 1) {
			scl_0; /* scl low */
			if (d & 0x80) { /* sda update */
                                sda_z;
			} else {
                                sda_0;
                        }
			scl_z; /* scl high */
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
                scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		if (ack) {
			fprintf(stderr, "rtc_present=0\n");
			goto skiprtc;
		}

		for (j = 0; j < 9; j++) {
			for (d = rtcdat[j], i = 0; i < 8; i++, d <<= 1) {
                                scl_0; /* scl low */
                                if (d & 0x80) { /* sda update */
                                        sda_z;
                                } else {
                                        sda_0;
                                }
                                scl_z; /* scl high */
			}
                        scl_0; /* scl low, tristate sda */
                        sda_z;
                        scl_z; /* scl high, tristate sda */
                        ack = sda_in; /* sample ack */
			assert(ack == 0);
		}

		scl_0; /* scl low, sda low */
                sda_0;
		scl_z; /* scl high, sda low */
		sda_z; /* tristate (i2c stop) */
		if (rtcdat[1] & 0x80) {
			rtcdat[1] &= ~0x80;
			goto resettime;
		}
		if (rtcdat[2] & 0x80) {
			sleep(6);
			rtcdat[2] &= ~0x80;
			goto resettime;
		}
		fprintf(stderr, "rtc_present=1\n");
	}

	if (opt_getrtc || opt_diffrtc) {
		unsigned int d, j, k, ack;
		//unsigned int ret;
		unsigned char rtcdat[9];
		time_t now;
		struct tm tm;
		struct timeval tv;
		//int cdrift, rdrift;
		//struct timex tmx;
                volatile unsigned int dummy;

		bzero(rtcdat, 9);
		//read_tagmem(tagmem);
		//cdrift = (signed char)((~tagmem[2] & 0xff00) >> 8) * 2;
		//rdrift = (signed char)(~tagmem[2] & 0xff);
		scl_z; /* tristate, scl/sda pulled up */
                sda_z;
		sda_0; /* i2c, start (sda low) */
		sda_z; /* i2c, stop (sda high) */
		sda_0; /* i2c, start (sda low) */
		for (d = 0xd0, i = 0; i < 8; i++, d <<= 1) {
			scl_0; /* scl low */
			if (d & 0x80) { /* sda update */
                                sda_z;
			} else {
                                sda_0;
                        }
			scl_z; /* scl high */
		}
                scl_0;
		sda_z; /* scl low, tristate sda */
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		printf("rtc_present=%d\n", ack ? 0 : 1);
	//	printf("tagmem_cpudrift_ppm=%d\n", cdrift);
	//	if (opt_getrtc) {
	//		tmx.modes = ADJ_FREQUENCY;
	//		tmx.freq = cdrift * 65536;
	//		ret = adjtimex(&tmx);
	//		assert(ret != -1);
	//	}
	//	printf("tagmem_rtcdrift_ppm=%d\n", rdrift);
		if (ack) goto skiprtc;
                sda_0;
		for (i = 0; i < 8; i++) {
			scl_0; /* scl low, sda low */
			scl_z; /* scl high */
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		assert(ack == 0);
		scl_0; /* scl low, tristate sda */
		scl_0; /* scl low, sda high */
		scl_z; /* scl high, sda high */
		sda_0; /* scl high, sda lo (repeated start) */
		for (d = 0xd1, i = 0; i < 8; i++, d <<= 1) {
			scl_0; /* scl low */
			if (d & 0x80) { /* sda update */
                                sda_z;
			} else {
                                sda_0;
                        }
			scl_z; /* scl high */
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		assert(ack == 0);
		for (j = 0; j < 8; j++) {
			for (i = 0; i < 8; i++) {
				scl_0; /* scl low, tri sda */
                                sda_z;
				scl_z; /* scl high, tri sda */
				scl_z; /* scl high, tri sda (for timing) */
				rtcdat[j] <<= 1;
				rtcdat[j] |= sda_in ? 1 : 0;
			}
			if (j != 7) {
				scl_0; /* scl low, tri sda */
				sda_0; /* scl low, sda low */
				scl_z; /* scl high, sda low */
			} else {
				scl_0; /* scl low, tri sda */
				sda_z; /* scl low, sda high */
				scl_z; /* scl high, sda hi */
			}
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
		sda_0; /* scl low, sda low */
		scl_z; /* scl high, sda low */
		sda_z; /* tristate (i2c stop) */
		printf("rtc_oscillator_ok=%d\n", (rtcdat[1] & 0x80) ? 0 : 1);
		printf("rtc_calibration_reg=0x%x\n", rtcdat[7]);
		printf("rtc_calibration_ppm=%f\n", rtcdat[7]&(1<<5) ?
		  ((rtcdat[7] & 0x1f) * 4.068) : ((rtcdat[7] & 0x1f) * -2.034));
		//for (i = 0; i < 8; i++) printf("dat%d=0x%x\n", i, rtcdat[i]);
		if ((rtcdat[1] & 0x80) == 0) {
			tm.tm_sec = frombcd(rtcdat[0] & 0x7f);
			tm.tm_min = frombcd(rtcdat[1] & 0x7f);
			tm.tm_hour = frombcd(rtcdat[2] & 0x3f);
			tm.tm_mday = frombcd(rtcdat[4] & 0x3f);
			tm.tm_mon = frombcd(rtcdat[5] & 0x1f) - 1;
			tm.tm_year = frombcd(rtcdat[6]) + 100;
			setenv("TZ", "UTC", 1);
			now = mktime(&tm);
			tv.tv_sec = now;
			tv.tv_usec = 0;
			if (opt_getrtc) settimeofday(&tv, NULL);
		}
	}

	if (opt_diffrtc) {
		struct sched_param sched;
		unsigned int d, j, k, ack;
		int loops = 0;
		unsigned char rtcdat[8], prev;
		time_t now;
		struct tm tm;
		struct timeval tv0, tv;
                volatile unsigned int dummy;

		bzero(&tm, sizeof(tm));
		sched.sched_priority = 99;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		setenv("TZ", "UTC", 1);
		gettimeofday(&tv0, NULL);

diffrtc_again:
		scl_z; /* tristate, scl/sda pulled up */
                sda_z;
		sda_0; /* i2c, start (sda low) */
		sda_z; /* i2c, stop (sda high) */
		sda_0; /* i2c, start (sda low) */
		for (d = 0xd0, i = 0; i < 8; i++, d <<= 1) {
			scl_0; /* scl low */
			if (d & 0x80) { /* sda update */
                                sda_z;
			} else {
                                sda_0;
                        }
			scl_z; /* scl high */
		}
                scl_0;
		sda_z; /* scl low, tristate sda */
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		assert(ack == 0);
                sda_0;
		for (i = 0; i < 8; i++) {
			scl_0; /* scl low, sda low */
			scl_z; /* scl high */
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		assert(ack == 0);
		scl_0; /* scl low, tristate sda */
		scl_0; /* scl low, sda high */
		scl_z; /* scl high, sda high */
		sda_0; /* scl high, sda lo (repeated start) */
		for (d = 0xd1, i = 0; i < 8; i++, d <<= 1) {
			scl_0; /* scl low */
			if (d & 0x80) { /* sda update */
                                sda_z;
			} else {
                                sda_0;
                        }
			scl_z; /* scl high */
		}
                scl_0;
		sda_z; /* scl low, tristate sda */
		scl_z; /* scl high, tristate sda */
		ack = sda_in; /* sample ack */
		assert(ack == 0);
		for (i = 0, prev = rtcdat[0]; i < 8; i++) {
                        scl_0; /* scl low, tri sda */
                        sda_z;
                        scl_z; /* scl high, tri sda */
                        scl_z; /* scl high, tri sda (for timing) */
			rtcdat[0] <<= 1;
			rtcdat[0] |= sda_in ? 1 : 0;
		}
		if (rtcdat[0] == prev || loops == 0) { /* no second hand movement? */
			scl_0; /* scl low, tri sda */
			scl_0; /* scl low, sda high */
			scl_z; /* scl high, sda hi */
			scl_0; /* scl low, tristate sda */
			sda_0; /* scl low, sda low */
			scl_z; /* scl high, sda low */
			sda_z; /* tristate (i2c stop) */
			loops++;
			assert(loops < 100000);
			goto diffrtc_again;
		} else {
			gettimeofday(&tv, NULL);
			scl_0; /* scl low, tri sda */
			sda_0; /* scl low, sda low */
			scl_z; /* scl high, sda low */
		}
		for (j = 1; j < 8; j++) {
			for (i = 0; i < 8; i++) {
				scl_0; /* scl low, tri sda */
                                sda_z;
				scl_z; /* scl high, tri sda */
				scl_z; /* scl high, tri sda (for timing) */
				rtcdat[j] <<= 1;
				rtcdat[j] |= sda_in ? 1 : 0;
			}
			if (j != 7) {
				scl_0; /* scl low, tri sda */
				sda_0; /* scl low, sda low */
				scl_z; /* scl high, sda low */
			} else {
				scl_0; /* scl low, tri sda */
				sda_z; /* scl low, sda high */
				scl_z; /* scl high, sda hi */
			}
		}
		scl_0; /* scl low, tristate sda */
                sda_z;
		sda_0; /* scl low, sda low */
		scl_z; /* scl high, sda low */
		sda_z; /* tristate (i2c stop) */
		//for (i = 0; i < 8; i++) printf("dat%d=0x%x\n", i, rtcdat[i]);
		if ((rtcdat[1] & 0x80) == 0) {
			long long us, p;
			tm.tm_sec = frombcd(rtcdat[0] & 0x7f);
			tm.tm_min = frombcd(rtcdat[1] & 0x7f);
			tm.tm_hour = frombcd(rtcdat[2] & 0x3f);
			tm.tm_mday = frombcd(rtcdat[4] & 0x3f);
			tm.tm_mon = frombcd(rtcdat[5] & 0x1f) - 1;
			tm.tm_year = frombcd(rtcdat[6]) + 100;
			fprintf(stderr, "tm=\"%d %d %d %d %d %d\"\n", tm.tm_sec,
			  tm.tm_min, tm.tm_hour, tm.tm_mday, tm.tm_mon, 
			  tm.tm_year);
			now = mktime(&tm);
			p = us = tv.tv_sec * 1000000LL + tv.tv_usec;
			us -= now * 1000000LL;
			fprintf(stderr, "diffrtc=");
			if (us < 0) fputc('-', stderr);
#define LLABS(x)	((x) < 0 ? -(x) : (x))
			fprintf(stderr, "%lld.%06lld\n", LLABS(us) / 1000000LL,
			  LLABS(us) % 1000000LL);
			fprintf(stderr, "measured_when=%lld\n", 
			  (long long)tv.tv_sec);
			fprintf(stderr, "loops=%d\n", loops);
			p -= tv0.tv_sec * 1000000LL + tv0.tv_usec;
			fprintf(stderr, "precision_us=%lld\n", p / loops);
			fprintf(stderr, "now=%lld\n", (long long)now);
			fprintf(stderr, "rtcdat=\"%02x", rtcdat[0]);
			for (i = 1; i < 7; i++) fprintf(stderr, " %02x", rtcdat[i]);
			fprintf(stderr, "\"\n");
		}
	}

skiprtc:
	misc_control_register_base[0x18 / 4] |= 0x00003000; /* pins back to I2C */


	/* Should be last! */
	if (opt_watchdog) {
		struct sched_param sched;
		int fd, r, grain = 0, n, mode, t;
		char c;
                unsigned int x = 0;
                fd_set rfds;
                struct timeval tv_now, tv_wait, tv_exp;

		daemon(0, 0);
		sched.sched_priority = 98;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		if (opt_autofeed) {
			unsigned int us = 60000000;
			if (wdog == 0) {
				us = 100000;
			} else if (wdog == 1) {
				us = 1000000;
			} else if (wdog == 2) {
				us = 5000000;
			}
			for (;;) {
				poke32(0xc, wdog);
				usleep(us);
			}
		}
#define FEED0 670000
#define FEED1 2680000
#define FEED2 10740000
		/* Single character writes to /dev/watchdog keep it fed */
                /* New /dev/watchdog interface:
                 * - Still supports the standard interface that expects
                 *   '0' '1' '2' or '3'.
                 * - Most other characters are still treated as 3.
                 * - 'f' (for feed) is a special character used to create
                 *   automatic feeds.
                 * - 'f' is expected to be followed by three digits, eg "f100".
                 * - Those digits are interpreted as a 3 digit number, which
                 *   is then divided by 10.  For example, "f456" feeds the
                 *   watchdog for 45.6 seconds.
                 * - An extended feed is canceled by any new write to
                 *   /dev/watchdog.
                 * - If 'f' is not followed by 3 digits, the behavior is
                 *   undefined, but in practice you typically just fed the
                 *   wdt for a long time.
                 */
		mknod("/dev/watchdog", S_IFIFO|0666, 0);
                mode = 0; // mode 1 == processing a multi-feed write
		while(1) {
                        fd = open("/dev/watchdog", O_RDONLY | O_NONBLOCK);
                        assert (fd != -1);
                        FD_ZERO(&rfds);
                        FD_SET(fd, &rfds);
                        if (mode)
                                n = select(fd + 1, &rfds, NULL, NULL, &tv_wait);
                        else
                                n = select(fd + 1, &rfds, NULL, NULL, NULL);
                        r = 0;
                        if (FD_ISSET(fd, &rfds)) r = read(fd, &c, 1);
                        if (r) { // data coming in /dev/watchdog
                                if (c == 'f') { // new style feed
                                        read(fd, &c, 1);
                                        t = (c - '0') * 100;
                                        read(fd, &c, 1);
                                        t += (c - '0') * 10;
                                        read(fd, &c, 1);
                                        t += (c - '0');
                                        t = t * 100000;
                                        x = 0;
                                        if (t > FEED1) x = 1;
                                        if (t > FEED2) x = 2;
                                        poke32(0xc, x);
                                        switch (x) {
                                                case 0: grain = FEED0; break;
                                                case 1: grain = FEED1; break;
                                                case 2: grain = FEED2; break;
                                        }
                                        mode = 1;
                                        gettimeofday(&tv_exp, NULL);
                                        t += tv_exp.tv_usec;
                                        tv_exp.tv_usec = t % 1000000;
                                        tv_exp.tv_sec += t / 1000000;
                                } else { // old style feed
                                        if (c > '3' || c < '0') c = '3';
                                        poke32(0xc, c - '0');
                                        mode = 0;
                                }
                        }
                        if (mode) {
                                // if in mode 1, we feed wdt and calculate
                                // how long to sleep
                                poke32(0xc, x);
                                gettimeofday(&tv_now, NULL);
                                t = tv_exp.tv_sec - tv_now.tv_sec;
                                t *= 1000000;
                                t += tv_exp.tv_usec;
                                t -= tv_now.tv_usec;
                                if (t <= grain) mode = 0;
                                else if (t <= 3*grain/2) t -= grain;
                                else t = grain/2;
                                tv_wait.tv_sec = t / 1000000;
                                tv_wait.tv_usec = t % 1000000;
                        }
                        // if mode was 0 and no character was read, don't do
                        // anything, just go back to waiting for a character
                        close(fd);
		} // end forever loop
	}

	munmap((void*)misc_control_register_base,  getpagesize());
	return 0;
}


